package main

import (
	"context"
	"crypto/tls"
	"fmt"
	"io"
	"log"
	"net/http"
	"os"
	"os/exec"
	"sync"

	"github.com/gin-gonic/gin"
	"gorm.io/gorm"
	"whiteboxsystems.nl/okapidemo/sharedmodel"
	"whiteboxsystems.nl/okapidemo/whiteboxservice/model"
)

type UIService struct {
	srv        *http.Server
	inited     bool
	clientCert tls.Certificate
	data       *gorm.DB
}

func (srv *UIService) LoadData(location string) error {
	var err error
	srv.data, err = model.GetDB(location)
	return err
}

func (srv *UIService) Addr() string {
	if srv.srv == nil {
		return ""
	}
	return srv.srv.Addr
}

func (srv *UIService) ListenAndServe() {
	if !srv.inited {
		srv.init()
	}
	log.Println("Listening on %v", srv.srv.Addr)
	srv.srv.ListenAndServe()
}

func (srv *UIService) Shutdown(ctx context.Context) error {
	return srv.srv.Shutdown(ctx)
}

func (srv *UIService) init() {
	r := srv.srv.Handler.(*gin.Engine)

	r.LoadHTMLGlob("templates/*")
	r.Static("/assets", "./assets")

	r.Use(srv.Authenticate)
	r.GET("/", func(c *gin.Context) {
		c.Redirect(301, "/ui")
	})
	r.GET("/ui", srv.GetIndex)
	r.GET("/ui/*page", srv.GetIndex)
	r.GET("/api/connections", srv.GetConnections)
	r.GET("/api/connections/:connID", srv.GetConnection)
	r.GET("/api/connections/:connID/:serviceID", srv.GetSubscriptions)
	r.GET("/api/connections/:connID/:serviceID/:patientID", srv.GetPatient)
	r.GET("/api/registrations", srv.GetRegistrations)
	// r.GET("/api/systems/:sysid/patients", srv.GetPatients)
	// r.GET("/api/systems/:sysid/patients/:patid", srv.GetPatient)
	srv.inited = true
}

func (srv *UIService) GetIndex(c *gin.Context) {
	c.HTML(http.StatusOK, "index.html", gin.H{})
}

func (srv *UIService) GetConnection(c *gin.Context) {
	connID := c.Param("connID")
	connection := &sharedmodel.Connection{}
	srv.data.Where("id = ?", connID).Find(&connection)
	c.JSON(200, connection)
}
func (srv *UIService) GetSubscriptions(c *gin.Context) {
	connID := c.Param("connID")
	serviceID := c.Param("serviceID")
	serviceConfig := &sharedmodel.ServiceConfig{}
	srv.data.Preload("Service").Preload("Subscriptions").Where("connection_id = ? and id = ?", connID, serviceID).Find(&serviceConfig)
	c.JSON(200, serviceConfig)
}

func (srv *UIService) GetPatient(c *gin.Context) {
	connID := c.Param("connID")
	serviceID := c.Param("serviceID")
	patientID := c.Param("patientID")
	patient := &sharedmodel.Subscription{}
	serviceConfig := &sharedmodel.ServiceConfig{}
	if err := srv.data.Preload("FetchProtocol").Preload("FetchProtocol.AuthConfig").Preload("Service").Where("connection_id = ? and id = ?", connID, serviceID).Find(&serviceConfig).Error; err != nil {
		c.AbortWithError(500, err)
		return
	}
	srv.data.Where("service_config_id = ? and id = ?", serviceID, patientID).Find(&patient)

	protoconfig := map[string]string{}
	protometa := map[string]string{}

	err := serviceConfig.FetchProtocol.UnmarshalConfig(&protoconfig)

	log.Println(err, protoconfig)

	err = patient.GetProtocolMeta(&protometa)

	log.Println(err, protometa)

	url := fmt.Sprintf("%v/%v/%v", protoconfig["url"], "patients", protometa["patientID"])

	req, _ := http.NewRequest("GET", url, nil)
	// req.Header.Set("Authorization", serviceConfig.FetchProtocol.AuthConfig.Raw)

	client := &http.Client{
		Transport: &http.Transport{
			TLSClientConfig: &tls.Config{
				InsecureSkipVerify: true,
				Certificates:       []tls.Certificate{srv.clientCert},
			},
		},
	}

	resp, err := client.Do(req)

	if err != nil {
		c.AbortWithError(500, err)
		return
	}

	if resp.StatusCode >= 400 {
		c.AbortWithError(resp.StatusCode, fmt.Errorf("%v", resp.Status))
		return
	}

	cmd := exec.Command(binFolder+"/ediviewer", binFolder+"/medeur.json", binFolder+"/template.html", "/dev/stdin")
	sin, err := cmd.StdinPipe()
	if err != nil {
		log.Println("[ediviewer] Failed to open stdin pipe:", err)
		c.AbortWithError(500, err)
		return
	}
	defer sin.Close()

	serr, err := cmd.StderrPipe()
	if err != nil {
		log.Println("[ediviewer] Failed to open stderr pipe:", err)
		c.AbortWithError(500, err)
		return
	}
	defer serr.Close()

	sout, err := cmd.StdoutPipe()
	if err != nil {
		log.Println("[ediviewer] Failed to open stdout pipe:", err)
		c.AbortWithError(500, err)
		return
	}
	defer sout.Close()

	if err := cmd.Start(); err != nil {
		log.Println("[ediviewer] Failed to start:", err)
		c.AbortWithError(500, err)
		return
	}

	wg := sync.WaitGroup{}
	wg.Add(3)

	go func() {
		defer wg.Done()
		if _, err := io.Copy(sin, resp.Body); err != nil {
			log.Println("[ediviewer] Error reading EDIFACT:", err)
			c.AbortWithError(500, err)
		}
	}()

	go func() {
		defer wg.Done()
		if _, err := io.Copy(c.Writer, sout); err != nil {
			log.Println("[ediviewer] Error output EDIFACT:", err)
			c.AbortWithError(500, err)
		}
	}()

	go func() {
		defer wg.Done()
		if _, err := io.Copy(os.Stderr, serr); err != nil {
			log.Println("[ediviewer] Error stderr EDIFACT:", err)
			c.AbortWithError(500, err)
		}
	}()

	wg.Wait()

	if err := cmd.Wait(); err != nil {
		log.Println("[ediviewer] Failed:", err)
		c.AbortWithError(500, err)
	}
}

func (srv *UIService) Authenticate(c *gin.Context) {
	// authHeader := c.Request.Header.Get("Authorization")
	// log.Printf("authHeader: %v", authHeader)

	// if authHeader != "1111" {
	// 	c.Status(401)
	// 	c.Abort()
	// }
}

func (srv *UIService) GetConnections(c *gin.Context) {
	connections := []*sharedmodel.Connection{}
	srv.data.Preload("Services").Preload("Services.Service").Preload("Services.Subscriptions").Find(&connections)
	c.JSON(200, connections)
}

func (srv *UIService) GetRegistrations(c *gin.Context) {
	registrations := []*sharedmodel.Registration{}
	srv.data.Where("status = ?", sharedmodel.RegistrationStatusPending).Find(&registrations)
	c.JSON(200, registrations)
}

// func (srv *UIService) GetSystems(c *gin.Context) {
// 	id := c.Param("id")
// 	for _, p := range patients {
// 		if p.PatientID == id {
// 			f, err := os.Open(path.Join("./data/patients", p.EDI))

// 			if err != nil {
// 				c.Error(err)
// 				return
// 			}

// 			io.Copy(c.Writer, f)
// 			return
// 		}
// 	}

// 	c.JSON(404, nil)
// }

// func (srv *UIService) GetPatients(c *gin.Context) {
// 	id := c.Param("id")
// 	for _, p := range patients {
// 		if p.PatientID == id {
// 			f, err := os.Open(path.Join("./data/patients", p.EDI))

// 			if err != nil {
// 				c.Error(err)
// 				return
// 			}

// 			io.Copy(c.Writer, f)
// 			return
// 		}
// 	}

// 	c.JSON(404, nil)
// }

// func (srv *UIService) GetPatient(c *gin.Context) {
// 	id := c.Param("id")
// 	for _, p := range patients {
// 		if p.PatientID == id {
// 			f, err := os.Open(path.Join("./data/patients", p.EDI))

// 			if err != nil {
// 				c.Error(err)
// 				return
// 			}

// 			io.Copy(c.Writer, f)
// 			return
// 		}
// 	}

// 	c.JSON(404, nil)
// }

func NewUIServer(addr string) *UIService {
	cert := loadCert()

	srv := &UIService{srv: &http.Server{
		Addr:    addr,
		Handler: gin.Default(),
	}, clientCert: *cert}

	return srv
}
